<?php
namespace Concrete\Package\DatabaseDumper\Database\Connection;

use Concrete\Package\DatabaseDumper\Utility\DB;
use Concrete\Package\DatabaseDumper\Utility\Str;

class Connection
{
    /**
     * @var PDO
     */
    private $db;

    private $loggedUser;

    private $charset;

    protected $dbname;

    public function __construct($driver)
    {
        $this->db = new PDO();
    }

    public function connect($server, $user, $pwd, $dbname)
    {
        $this->dbname = $dbname;
        if ($this->db->connect($server, $user, $pwd)) {
            $this->db->select_db($dbname);
            $this->db->set_charset($this->charset = ($this->min_version("5.5.3", 0) ? "utf8mb4" : "utf8"));
            $this->db->query("SET sql_quote_show_create = 1, autocommit = 1");

            return true;
        }

        return false;
    }

    public function slowQuery($query, $timeout) {
        if ($this->min_version('5.7.8', '10.1.2')) {
            if (preg_match('~MariaDB~', $this->db->server_info)) {
                return "SET STATEMENT max_statement_time=$timeout FOR $query";
            } elseif (preg_match('~^(SELECT\b)(.+)~is', $query, $match)) {
                return "$match[1] /*+ MAX_EXECUTION_TIME(" . ($timeout * 1000) . ") */ $match[2]";
            }
        }
    }

    public function result($statment, $colIdx = 0)
    {
        return $this->db->result($statment, $colIdx);
    }

    public function query($sql, $_g = false)
    {
        return $this->db->query($sql, $_g);
    }

    public function getVals($sql, $colIdx = 0)
    {
        $val = [];
        $r = $this->db->query($sql);
        if (is_object($r)) {
            while ($row = $r->fetch_row()) {
                $val[] = $row[$colIdx];
            }
        }

        return $val;
    }

    public function get_key_vals($sql, $set_keys = true)
    {
        $r = $this->db->query($sql);
        $return = [];
        if (is_object($r)) {
            while ($row = $r->fetch_row()) {
                if ($set_keys) {
                    $return[$row[0]] = $row[1];
                } else {
                    $return[] = $row[0];
                }
            }
        }

        return $return;
    }

    public function get_rows($sql)
    {
        $rows = [];
        $r = $this->db->query($sql);
        if (is_object($r)) {
            while ($row = $r->fetch_assoc()) {
                $rows[] = $row;
            }
        }

        return $rows;
    }

    public function support($feature)
    {
        return !preg_match("~scheme|sequence|type|view_trigger|materializedview" . ($this->min_version(8) ? "" : "|descidx" . ($this->min_version(5.1) ? "" : "|event|partitioning" . ($this->min_version(5) ? "" : "|routine|trigger|view"))) . "~", $feature);
    }

    public function fk_support($r)
    {
        return preg_match('~InnoDB|IBMDB2I~i', $r["Engine"]) || (preg_match('~NDB~i', $r["Engine"]) && $this->min_version(5.6));
    }

    /** Get tables list
     * @return array array($name => $type)
     */
    public function tables_list()
    {
        $result = $this->db->query("SELECT TABLE_NAME AS Name, TABLE_TYPE AS Type, ENGINE AS Engine, TABLE_COMMENT AS Comment FROM information_schema.TABLES WHERE TABLE_SCHEMA = {$this->getDBName(true)} ORDER BY Name");
        $tables = [];
        if (is_object($result)) {
            $tables = $result->fetchAll();
        }

        return $tables;
    }

    /** Get information about fields
     * @param string
     *
     * @return array array($name => array("field" => , "full_type" => , "type" => , "length" => , "unsigned" => , "default" => , "null" => , "auto_increment" => , "on_update" => , "collation" => , "privileges" => , "comment" => , "primary" => ))
     */
    public function fields($table)
    {
        $return = [];
        foreach ($this->get_rows("SHOW FULL COLUMNS FROM " . DB::table($table)) as $row) {
            preg_match('~^([^( ]+)(?:\((.+)\))?( unsigned)?( zerofill)?$~', $row["Type"], $match);
            $return[$row["Field"]] = [
                "field" => $row["Field"],
                "full_type" => $row["Type"],
                "type" => $match[1],
                "length" => $match[2],
                "unsigned" => ltrim($match[3] . $match[4]),
                "default" => ("" != $row["Default"] || preg_match("~char|set~", $match[1]) ? $row["Default"] : null),
                "null" => ("YES" == $row["Null"]),
                "auto_increment" => ("auto_increment" == $row["Extra"]),
                "on_update" => (preg_match('~^on update (.+)~i', $row["Extra"], $match) ? $match[1] : ""), //! available since MySQL 5.1.23
                "collation" => $row["Collation"],
                "privileges" => array_flip(preg_split('~, *~', $row["Privileges"])),
                "comment" => $row["Comment"],
                "primary" => ("PRI" == $row["Key"]),
            ];
        }

        return $return;
    }

    /** Get database collation
     * @param string
     * @param array result of collations()
     *
     * @return string
     */
    public function db_collation($db, $collations)
    {
        $return = null;
        $create = $this->db->result("SHOW CREATE DATABASE " . DB::table($db), 1);
        if (preg_match('~ COLLATE ([^ ]+)~', $create, $match)) {
            $return = $match[1];
        } elseif (preg_match('~ CHARACTER SET ([^ ]+)~', $create, $match)) {
            // default collation
            $return = $collations[$match[1]][-1];
        }

        return $return;
    }

    /** Get SQL command to create table
     * @param string
     * @param bool
     *
     * @return string
     */
    public function create_sql($table, $auto_increment)
    {
        $return = $this->db->result("SHOW CREATE TABLE " . DB::table($table), 1);
        if (!$auto_increment) {
            $return = preg_replace('~ AUTO_INCREMENT=\d+~', '', $return); //! skip comments
        }

        return $return;
    }

    /** Get select clause for convertible fields
     * @param array
     * @param array
     * @param array
     *
     * @return string
     */
    public function convert_fields($columns, $fields, $select = [])
    {
        $return = "";
        foreach ($columns as $key => $val) {
            $escapedKey = Str::idf_escape($key);
            if ($select && !in_array($escapedKey, $select)) {
                continue;
            }
            $as = $this->convert_field($fields[$key]);
            if ($as) {
                $return .= ", $as AS " . $escapedKey;
            }
        }

        return $return;
    }

    /** Convert field in select and edit
     * @param array one element from fields()
     *
     * @return string
     */
    public function convert_field($field)
    {
        if (preg_match("~binary~", $field["type"])) {
            return "HEX(" . Str::idf_escape($field["field"]) . ")";
        }
        if ("bit" == $field["type"]) {
            return "BIN(" . Str::idf_escape($field["field"]) . " + 0)"; // + 0 is required outside MySQLnd
        }
        if (preg_match("~geometry|point|linestring|polygon~", $field["type"])) {
            return ($this->min_version(8) ? "ST_" : "") . "AsWKT(" . Str::idf_escape($field["field"]) . ")";
        }
    }

    /** Convert value in edit after applying functions back
     * @param array one element from fields()
     * @param string
     *
     * @return string
     */
    public function unconvert_field($field, $return)
    {
        if (preg_match("~binary~", $field["type"])) {
            $return = "UNHEX($return)";
        }
        if ("bit" == $field["type"]) {
            $return = "CONV($return, 2, 10) + 0";
        }
        if (preg_match("~geometry|point|linestring|polygon~", $field["type"])) {
            $return = ($this->min_version(8) ? "ST_" : "") . "GeomFromText($return, SRID($field[field]))";
        }

        return $return;
    }

    /** Get SQL commands to create triggers
     * @param string
     * @param string
     *
     * @return string
     */
    public function trigger_sql($table)
    {
        $return = "";
        $rows = $this->get_rows("SHOW TRIGGERS LIKE " . $this->quote(addcslashes($table, "%_\\")), null, "-- ");
        foreach ($rows as $row) {
            $return .= "\nCREATE TRIGGER " . Str::idf_escape($row["Trigger"]) . " $row[Timing] $row[Event] ON " . DB::table($row["Table"]) . " FOR EACH ROW\n$row[Statement];;\n";
        }

        return $return;
    }

    public function logged_user()
    {
        if (!$this->loggedUser) {
            $this->loggedUser = $this->db->result("SELECT USER()");
        }

        return $this->loggedUser;
    }

    public function getCharset()
    {
        return $this->charset;
    }

    public function quote($s)
    {
        return $this->db->quote($s);
    }

    public function getDBName($quote = false)
    {
        return $quote ? $this->quote($this->dbname) : $this->dbname;
    }

    public function error()
    {
        return $this->db->error;
    }

    /** Check if connection has at least the given version
     * @param string required version
     * @param string required MariaDB version
     *
     * @return bool
     */
    public function min_version($version, $maria_db = "")
    {
        $server_info = $this->db->server_info;
        if ($maria_db && preg_match('~([\d.]+)-MariaDB~', $server_info, $match)) {
            $server_info = $match[1];
            $version = $maria_db;
        }

        return version_compare($server_info, $version) >= 0;
    }
}
