<?php
namespace Concrete\Package\DatabaseDumper;

use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Concrete\Package\DatabaseDumper\Http\FileResponse;
use Concrete\Package\DatabaseDumper\System\FileManager;
use Concrete\Package\DatabaseDumper\Database\Connection\Connection;
use Concrete\Package\DatabaseDumper\Utility\Str;
use Concrete\Package\DatabaseDumper\Utility\DB;

class Dumper
{
    /**
     * @var FileManager
     */
    private $fm;

    /**
     * @var string
     */
    private $outputDir;

    /**
     * @var string
     */
    private $outputFileName;

    /**
     * @var array
     */
    protected $options;

    use ApplicationTrait;

    public function __construct()
    {
        $fh = $this->app('helper/file');
        $this->outputFileName = $fh->sanitize($this->getSiteName()) . '.sql';
        $this->outputDir = 'dmp-' . time();
        $this->fm = new FileManager($fh->getTemporaryDirectory());
    }

    /**
     * getSiteName.
     *
     * @return string
     */
    private function getSiteName()
    {
        if ($this->app()->bound('site')) {
            return $this->app('site')->getSite()->getSiteName();
        }

        return $this->app('config')->get('concrete.site');
    }

    /**
     * @param array $options
     */
    public function run($options = [])
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $this->setOptions($options);

        if (function_exists("mb_internal_encoding")) {
            @mb_internal_encoding("8bit");
        }

        $this->dump('START');

        $dump = $this->dumpRoutines() . $this->dumpEvents();
        if (!empty($dump)) {
            $this->dump("DELIMITER ;;\n\n$dump" . "DELIMITER ;\n\n");
            unset($dump);
        }

        $is_sql = "sql" == $this->options['format'];
        $views = [];
        $tables = $db->tables_list();
        foreach ($tables as $table) {
            $table['fields'] = $db->fields($table['Name']);
            $this->dump($this->dumpTableStructure($table, DB::is_view($table) ? 2 : 0));
            if (DB::is_view($table)) {
                $views[] = $table['Name'];
            } else {
                $selectQuery = "SELECT *" . $db->convert_fields($table['fields'], $table['fields']) . " FROM " . DB::table($table['Name']);
                $this->dump($this->dumpTableData($table, $selectQuery));
            }
            if ($is_sql && ($triggers = $db->trigger_sql($table['Name']))) {
                $this->dump("\nDELIMITER ;;\n$triggers\nDELIMITER ;\n");
            }
            if ($is_sql) {
                $this->dump("\n");
            }
        }

        foreach ($views as $view) {
            $this->dump($this->dumpTableStructure($view, 1));
        }

        $this->dump('END');
        $this->download();
    }

    public function dumpRoutines()
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $routines = '';
        foreach (["FUNCTION", "PROCEDURE"] as $routine) {
            $rows = $db->get_rows("SHOW $routine STATUS WHERE Db = " . $db->getDBName(true));
            foreach ($rows as $row) {
                $create = DB::remove_definer($db->result("SHOW CREATE $routine " . Str::idf_escape($row["Name"]), 2), $db->logged_user());
                $routines .= ('DROP+CREATE' != $this->options['table_style'] ? "DROP $routine IF EXISTS " . Str::idf_escape($row["Name"]) . ";;\n" : "") . "$create;;\n\n";
            }
        }

        return $routines;
    }

    public function dumpEvents()
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $events = '';
        $rows = $db->get_rows("SHOW EVENTS", null, "-- ");
        foreach ($rows as $row) {
            $create = DB::remove_definer($db->result("SHOW CREATE EVENT " . Str::idf_escape($row["Name"]), 3));
            $events .= ('DROP+CREATE' != $this->options['table_style'] ? "DROP EVENT IF EXISTS " . Str::idf_escape($row["Name"]) . ";;\n" : "") . "$create;;\n\n";
        }

        return $events;
    }

    /** Export table structure
     * @param string
     * @param int 0 table, 1 view, 2 temporary view table
     *
     * @return string table structure
     */
    public function dumpTableStructure($table, $isView = 0)
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $tableStructure = '';

        if (2 == $isView) {
            $fields = [];
            foreach ($table['fields'] as $name => $field) {
                $fields[] = Str::idf_escape($name) . " $field[full_type]";
            }
            $create = "CREATE TABLE " . DB::table($table['Name']) . " (" . implode(", ", $fields) . ")";
        } else {
            $create = $db->create_sql($table['Name'], $this->options['auto_increment']);
        }

        if ($create) {
            if ("DROP+CREATE" == $this->options['table_style'] || 1 == $isView) {
                $tableStructure = "DROP " . (2 == $isView ? "VIEW" : "TABLE") . " IF EXISTS " . DB::table($table['Name']) . ";\n";
            }
            if (1 == DB::is_view($table)) {
                $create = DB::remove_definer($create, $db->logged_user());
            }
            $tableStructure .= "$create;\n\n";
        }

        return $tableStructure;
    }

    /** Export table data
     * @param string
     * @param string
     *
     * @return string
     */
    public function dumpTableData($table, $query)
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $max_packet = 1048576; // default, minimum is 1024

        $result = $db->query($query, 1); // 1 - MYSQLI_USE_RESULT //! enum and set as numbers
        if ($result) {
            $insert = "";
            $dump = "";
            $buffer = "";
            $keys = [];
            $suffix = "";
            $fetch_function = ($table != '' ? 'fetch_assoc' : 'fetch_row');
            while ($row = $result->$fetch_function()) {
                if (empty($keys)) {
                    $values = [];
                    foreach ($row as $val) {
                        $field = $result->fetch_field();
                        $keys[] = $field->name;
                        $key = Str::idf_escape($field->name);
                        $values[] = "$key = VALUES($key)";
                    }
                    $suffix = ";\n";
                }

                if (!$insert) {
                    $insert = "INSERT INTO " . DB::table($table['Name']) . " (" . implode(", ", array_map([Str::class, 'idf_escape'], $keys)) . ") VALUES";
                }

                foreach ($row as $key => $val) {
                    $field = $table['fields'][$key];
                    $row[$key] = ($val !== null
                        ? $db->unconvert_field($field, preg_match('((?<!o)int(?!er)|numeric|real|float|double|decimal|money)', $field["type"]) && !preg_match('~\[~', $field["full_type"]) && is_numeric($val) ? $val : $db->quote(($val === false ? 0 : $val)))
                        : "NULL"
                    );
                }
                $s = ($max_packet ? "\n" : " ") . "(" . implode(", ", $row) . ")";
                if (!$buffer) {
                    $buffer = $insert . $s;
                } elseif (strlen($buffer) + 4 + strlen($s) + strlen($suffix) < $max_packet) { // 4 - length specification
                    $buffer .= ",$s";
                } else {
                    $dump .= $buffer . $suffix;
                    $buffer = $insert . $s;
                }
            }

            if ($buffer) {
                $dump .= $buffer . $suffix;
            }

            return $dump;
        } else {
            return "-- " . str_replace("\n", " ", $db->error()) . "\n";
        }
    }

    private function dump($data)
    {
        if ('START' === $data) {
            $this->fm->createFile($this->outputFile());
            $data = $this->dumpSqlHeader();
        } elseif ('END' === $data) {
            $data = $this->dumpSqlFooter();
        }

        $this->fm->appendToFile($this->outputFile(), $data);
    }

    public function dumpSqlHeader()
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');
        $db->query("SET time_zone = '+00:00';");

        return "SET FOREIGN_KEY_CHECKS=0;\n" .
               "SET SQL_MODE = \"NO_AUTO_VALUE_ON_ZERO\";\n" .
               "SET time_zone = \"+00:00\";\n\n" .
               "/*!40101 SET NAMES {$db->getCharset()} */;\n\n";
    }

    public function dumpSqlFooter()
    {
        /* @var Connection $db */
        $db = $this->app('xan/dumper/db');

        return "\n\nSET FOREIGN_KEY_CHECKS=1;\n" .
               "\n-- " . $db->result("SELECT NOW()") . "\n";
    }

    protected function setOptions($options = [])
    {
        $this->options = $options + [
            'output' => 'file',
            'format' => 'sql',
            'auto_increment' => true,
            'table_style' => 'DROP+CREATE',
         ];
    }

    private function download()
    {
        $this->app('session')->save();

        list($outputFile, $outputFileName, $contentType) = $this->prepareDownloadFormat();

        $response = new FileResponse($this->fm->getFullPath($outputFile));
        $response->setContentDisposition(ResponseHeaderBag::DISPOSITION_ATTACHMENT, $outputFileName);
        $response->headers->set('Content-Type', $contentType);
        $response->prepare();
        $response->send();

        $this->cleanUpTempFiles();
        $this->app()->shutdown();
    }

    private function prepareDownloadFormat()
    {
        $outputFile = $this->outputFile();
        $outputFileName = $this->outputFileName;
        $contentType = 'application/octet-stream';

        if ('zip' == $this->options['output']) {
            $outputFile = ($outputFileName .= '.zip');
            $contentType = 'application/zip';
            $this->fm->zip($this->outputDir, $outputFileName);
        } elseif ('gzip' == $this->options['output']) {
            $outputFile = ($outputFileName .= '.gz');
            $contentType = 'application/x-download';
            $this->fm->gzip($this->outputFile(), $outputFileName);
        }

        return [$outputFile, $outputFileName, $contentType];
    }

    private function outputFile()
    {
        return "{$this->outputDir}/{$this->outputFileName}";
    }

    private function cleanUpTempFiles()
    {
        $this->fm->deleteDir($this->outputDir);
        $this->fm->deleteIfExists("{$this->outputFileName}.zip");
        $this->fm->deleteIfExists("{$this->outputFileName}.gz");
    }

    public function __destruct()
    {
        if (is_object($this->fm)) {
            $this->cleanUpTempFiles();
        }
    }
}
