<?php
namespace Concrete\Package\DatabaseDumper\System;

use League\Flysystem\Config;
use League\Flysystem\Filesystem;
use League\Flysystem\Adapter\Local as LocalAdapter;
use Concrete\Package\DatabaseDumper\ApplicationTrait;
use Concrete\Core\File\Service\Zip;

class FileManager
{
    /**
     * @var string
     */
    private $basePath;

    /**
     * @var Filesystem
     */
    private $fs;

    use ApplicationTrait;

    public function __construct($basePath)
    {
        $this->basePath = $basePath;
        if (!ends_with($this->basePath, ['/', '\\'])) {
            $this->basePath .= '/';
        }
        $this->fs = new Filesystem(
            new LocalAdapter($this->basePath, FILE_APPEND),
            new Config(['disable_asserts' => true, 'visibility' => LocalAdapter::VISIBILITY_PRIVATE])
        );
    }

    /**
     * Get Full path of the given path.
     *
     * @param string $path
     *
     * @return string Full path
     */
    public function getFullPath($path)
    {
        return $this->basePath . ltrim($path, '\\/');
    }

    /**
     * Check whether a file exists.
     *
     * @param string $path
     *
     * @return bool
     */
    public function has($path)
    {
        return $this->fs->has($path);
    }

    /**
     * Create a directory.
     *
     * @param string $dirname directory name
     *
     * @return array|false
     */
    public function createDir($dirname)
    {
        return $this->fs->createDir($dirname);
    }

    /**
     * Create a new file.
     *
     * @param string $path
     *
     * @return bool
     */
    public function createFile($path)
    {
        return $this->fs->write($path, '');
    }

    /**
     * Update a file.
     *
     * @param string $path
     * @param string $content
     *
     * @return bool
     */
    public function appendToFile($path, $content)
    {
        return $this->fs->update($path, $content);
    }

    /**
     * Delete a file if exists.
     *
     * @param string $path
     *
     * @return bool
     */
    public function deleteIfExists($path)
    {
        if ($this->has($path)) {
            return $this->delete($path);
        }

        return false;
    }

    /**
     * Delete a file.
     *
     * @param string $path
     *
     * @return bool
     */
    public function delete($path)
    {
        return $this->fs->delete($path);
    }

    /**
     * Delete a directory.
     *
     * @param string $dirname
     *
     * @return bool
     */
    public function deleteDir($dirname)
    {
        return $this->fs->deleteDir($dirname);
    }

    /**
     * Compress the contents of a directory to a ZIP archive.
     *
     * @param string $sourceDirectory the directory to compress
     * @param string $zipFile the ZIP file to create (it will be deleted if already existing, unless the 'append' option is set to true)
     *
     * @throws \Exception
     */
    public function zip($sourceDirectory, $zipFile)
    {
        $this->app(Zip::class)->zip($this->getFullPath($sourceDirectory), $this->getFullPath($zipFile));
    }

    /**
     * Compress file to a GZIP file.
     *
     * @param string $sourceFile the file to compress
     * @param string $gzipFile the GZIP file to create
     */
    public function gzip($sourceFile, $gzipFile)
    {
        $gzdata = gzencode(implode("", file($this->getFullPath($sourceFile))), 9);
        $this->fs->write($gzipFile, $gzdata);
    }
}
